(* 
 File: Parser.sml

 The structure which contains the PEPA Nets parser
*)

structure Parser :> Parser =
struct

(* Parser tools.  Taken from Chris Reade's
   book ``Elements of Functional Programming''.
*)

(* Utility functions *) 

fun pair a b = (a,b)
fun fst(x,y) = x
fun snd(x,y) = y

fun consonto x a = a :: x
fun fold f [] b     = b
  | fold f (h::t) b = f (h, fold f t b)
fun link llist = fold (op @) llist []

datatype 'a possible = 
    Ok of 'a 
  | Fail

type 'a parser = Lexer.token list -> ('a * Lexer.token list) possible

infixr 4 <&>
infixr 3 <|>
infix  0 modify
 
fun (parser1 <|> parser2) s =
    let fun parser2_if_fail Fail = parser2 s
          | parser2_if_fail x    = x
    in
        parser2_if_fail (parser1 s)
    end

fun (parser modify f) s =
    let fun modresult Fail        = Fail
	  | modresult (Ok (x, y)) = Ok (f x, y)
    in
        modresult (parser s)
    end

fun (parser1 <&> parser2) s =
    let fun parser2_after Fail          = Fail
	  | parser2_after (Ok (x1, s1)) = (parser2 modify (pair x1)) s1
    in
        parser2_after (parser1 s)
    end
 
fun emptyseq s = Ok ([], s)

fun optional pr = (pr modify (consonto []))
                  <|> emptyseq

fun sequence pr =
    let fun seqpr s = ((pr <&> seqpr modify (op ::))
                       <|> emptyseq) s
    in
        seqpr
    end

fun seqwith (front, sep, back) pr =
    let val sep_pr = sep <&> pr               modify snd
        val items  = pr  <&> sequence sep_pr  modify (op ::)	
    in
	front <&> optional items <&> back modify (link o fst o snd)
    end

fun parserList []           = emptyseq
  | parserList (pr :: rest) = pr <&> (parserList rest) modify (op ::)

fun alternatives []           = (fn x => Fail)
  | alternatives (pr :: rest) = pr <|> alternatives rest;
    

 
(* Basic parsers *)

fun variable (Lexer.Ident x :: s)   = Ok (x, s)
  | variable other                  = Fail 

fun float (Lexer.Float x :: s)      = Ok (x, s)
  | float other                     = Fail
                                                                                
fun literal a (Lexer.Symbol x :: s) = if a = x then Ok (x, s) else Fail
  | literal a other                 = Fail;
    


(* A parser for PEPA.
*)


fun unparenth (bra, (e, ket)) = e;
    
fun parenseq s = seqwith (literal #"(", literal #",", literal #")") s;

fun bracedseq s = seqwith (literal #"{", literal #",", literal #"}") s;

fun angledseq s = seqwith (literal #"<", literal #",", literal #">") s;

val actlist : PepaNets.Identifier list parser    = bracedseq variable;

val cooplist: PepaNets.Identifier list parser    = angledseq variable;

(* val componentList: PepaNets.Component list parser = parenseq agenta; *)

fun def  s   = ((adef <&> optional (literal #";" <&> def) modify opt_consts)
               <|> marking                                                   ) s

and marking s = ((parenseq agenta) modify mk_marking) s

and adef s   = ((optional (literal #"#")  <&> variable  <&>
                literal #"="     <&> agenta                 modify mk_const)
               <|> arc                                                    
               <|> (priorities                               modify mk_prior)) s 

and arc s    = ((variable <&> literal #"-" <&> literal #"(" <&> variable
                          <&> literal #"," <&> variable <&> literal #")" 
                          <&> literal #"-" <&> literal #">" <&> variable)
                                                              modify mk_arc) s

and priorities s = 
               (((varList <&> literal #">" <&> priorities) modify mk_priorities)
               <|> (varList                                modify mk_priorities')) s

and varList s = (((variable <&> literal #"," <&> varList)    modify mk_varList')
                <|> (variable                                 modify mk_varList)) s

and agenta s = (agentb <&> optional (literal #"/" <&> actlist) 
                                                           modify mk_hiding) s

and agentb s = (agentc <&> optional (literal #"+" <&> agentb) 
                                                             modify mk_plus) s

and agentc s = (agentd <&> optional (cooplist <&> agentc) 
                                                             modify mk_coop) s

and agentd s = (agent <&> optional (literal #"[" <&> cellFiller <&> literal #"]") 
                                                             modify mk_cell) s

and agent s  = ((literal #"("    <&> variable    <&>
                 literal #","    <&> variable    <&>
                 literal #")"    <&> literal #"." <&> 
                 agent                                    modify mk_prefix)
               <|> (literal #"(" <&> agenta      <&> literal #")" 
                                                          modify unparenth)
               <|> (variable                              modify mk_agent)     
               <|> (float                                 modify mk_rate_literal)) s

and cellFiller s = ((literal #"_"            modify mk_null_cell)
                <|> (agent                   modify mk_some_cell)) s

and mk_agent s                               = PepaNets.VAR s

and mk_rate_literal s                        = PepaNets.RATE s

and mk_const (hash, (i, (eq, e)))            = (i, e)

and mk_hiding (P, [])                        = P
  | mk_hiding (P, [(slash, L)])              = PepaNets.HIDING (P, L)
  | mk_hiding (P, _)                         = error ("hiding", P)

and mk_plus (P, [])                          = P
  | mk_plus (P, [(plus, Q)])                 = PepaNets.CHOICE (P, Q)
  | mk_plus (P, _)                           = error ("plus", P)

and mk_coop (P, [])                          = P
  | mk_coop (P, [(L, Q)])                    = PepaNets.COOP ([P, Q], L)
  | mk_coop (P, _)                           = error ("cooperation", P)

and mk_cell (P, [])                          = P
  | mk_cell (P as PepaNets.VAR I, [(bra, (P', ket))])
                                             = PepaNets.CELL (I, P')
  | mk_cell (P, _)                           = error ("cell", P)

and mk_marking L                             = PepaNets.MARKING L

and mk_prefix (bra, (alpha, (comma, (rate, (ket, (dot, P))))))
                                             = PepaNets.PREFIX ((alpha, mk_rate rate), P)

and mk_prior p                               = (HashTable.hash "_PRIORITIES", PepaNets.PRIORITIES p)
and mk_priorities  (vars, (gt, priorities))  = vars :: priorities
and mk_priorities' (vars)                    = [vars]

and mk_varList (var)                         = [var]
and mk_varList' (var, (comma, varList))      = var:: varList

and mk_arc (P1, (hyphen, (bra, (alpha, (comma, (rate, (ket, (hyphen2, (gt, P2)))))))))
                                             = (HashTable.hash "_ARC", PepaNets.ARC (P1, (alpha, mk_rate rate), P2))

and mk_rate r                                = if HashTable.isInfty r 
                                               then PepaNets.Top else PepaNets.Rate r

and mk_null_cell _                           = NONE

and mk_some_cell P                           = SOME P

and opt_consts ((i, e1), [(oper, e2)])       = PepaNets.CONSTS (i, e1, e2)
  | opt_consts ((i, P), _)                   = error ("constants", P)

and error (s, P)                             = (Prettyprinter.setMode Prettyprinter.uncompressed;
                                                Error.parse_error ("[" ^ s ^ "] at or near:\n\t" ^ 
                                                                   Prettyprinter.print P))
     
local
  fun lit (Lexer.Symbol #";") = ";\n\t"
    | lit (Lexer.Symbol #"=") = " = "
    | lit (Lexer.Symbol #"+") = " + "
    | lit (Lexer.Symbol #",") = ", "
    | lit (Lexer.Symbol #"<") = " <"
    | lit (Lexer.Symbol #">") = "> "
    | lit (Lexer.Symbol c) = str c
    | lit (Lexer.Float f)  = f
    | lit (Lexer.Ident s)  = HashTable.unhash s
in
  fun report Fail         = Error.parse_error "ill-formed PEPA model definition"
    | report (Ok (c, [])) = c
    | report (Ok (c, x))  = Error.parse_error (String.concat 
                                  ("Unparsed :-\n" ::  (map lit x)))
end

(* Only for debugging *)
fun show C = (Prettyprinter.setMode Prettyprinter.verbose;
              TextIO.output (TextIO.stdOut, Prettyprinter.print C ^ "\n"); C)

val parse = report o def

end;
