(* 
 File: PepaNets.sml

 Datatypes of the PEPA Workbench [for PEPA Nets].
*)

structure PepaNets :> PepaNets =
struct
  type Identifier = HashTable.hashedstring 

  type Name = int

  datatype Rate = 
          Top
        | Rate of Identifier
        | Product of int * Identifier 
 
  datatype Component =  
	  PREFIX     of (Identifier * Rate) * Component
	| CHOICE     of Component * Component
	| COOP       of Component list * Identifier list
	| HIDING     of Component * Identifier list
        | CELL       of Identifier * Component option
	| VAR        of Identifier
        | RATE       of string
	| CONSTS     of Identifier * Component * Component
	| ARC        of Identifier * (Identifier * Rate) * Identifier
	| PRIORITIES of Identifier list list
	| MARKING    of Component list

  type State = Component

  type Environment = (Identifier * Component) list

  type Transition = Component * 
       ((Identifier * Rate) * Name) option * Component

  (* This function is exported from this structure *)
  (* Determine the minimum of two rates *)
  fun min (s, Top) = s
    | min (Top, s) = s
    | min (Rate s, Rate t) =
          if s = t then Rate s 
          else Rate (HashTable.literal("min(" ^ HashTable.show s ^ "," ^ HashTable.show t ^ ")"))
    | min (Product(m, s), Rate t) =
          if m = 1 then min (Rate s, Rate t)
          else Rate (HashTable.literal("min(" ^ Int.toString m ^ "*" ^ HashTable.show s ^ "," ^ 
                                                HashTable.show t ^ ")"))
    | min (Rate s, Product(n, t)) =
          if n = 1 then min (Rate s, Rate t)
          else Rate (HashTable.literal("min(" ^ HashTable.show s ^ "," ^ 
                                                Int.toString n ^ "*" ^ HashTable.show t ^ ")"))
    | min (Product(m, s), Product(n, t)) =
          if m = n then min (Rate s, Rate t)
          else Rate (HashTable.literal("min(" ^ Int.toString m ^ "*" ^ HashTable.show s ^ "," ^ 
                                                Int.toString n ^ "*" ^ HashTable.show t ^ ")"))

(* 
   Identifiers and Names 
   =====================
   An identifier is an alphanumeric descriptor assigned by the PEPA modeller
   whereas a name is a numeric stamp computed by the PEPA Workbench to allow
   the workbench to distinguish between repeated occurrences of a component.
   Names are assigned by the `derivative' function on traversing the tree
   of the abstract syntax of the PEPA model.  Names are assigned thus:

                               0
                              / \
                             1   2
                            / \ / \
                           3  4 5  6
   
   doubling as we go down the tree and adding 1 if we go left and 2 if we go 
   right.  This will make a distinction between P <> P <> P <> P, numbering the
   copies 1, 5, 13 and 14 and a bushier tree, (P <> P) <> (P <> P), numbering
   these copies 3, 4, 5, 6.  In either case the effect is the same: to have
   unique numbers for the leaves of the tree (the sequential components) and 
   the nodes (the cooperations).
*)


end;
