(* 
 File: Prettyprinter.sml

 The structure which prettyprints PEPA models
*)

structure Prettyprinter :> Prettyprinter =
struct

  open PepaNets

  datatype printMode = compressed | uncompressed | verbose

  val mode = ref compressed

  fun setMode m = mode := m

  fun print (PREFIX ((alpha, rate), P)) 
    = "("^printId alpha^", "^printRate rate^")."^(print P)
    | print (CHOICE (P, Q)) 
      = (print P) 
	^ (if !mode = verbose then "\n    + " else " + ")
        ^ (print Q)
    | print (COOP ([P, Q], L)) 
      = (if !mode = verbose then "(" else "")
	^ (print P) 
	^ (if !mode = verbose then " <" ^ (printList L) ^ "> " else "|")
	^ (print Q) 
        ^ (if !mode = verbose then ")" else "")
    | print (COOP (h :: t, L)) 
      = "(" ^ (printlist (h :: t)) ^ ")_{"
	^ (if !mode = verbose then " " ^ (printList L) ^ " " else ".")
        ^ "} "  
    | print (COOP ([], L)) = Error.fatal_error ("non-well-formed empty co-operation")
    | print (HIDING (P, [])) 
      = (print P)
    | print (HIDING (P, L)) 
      = (print P) ^ "/{" ^ (printList L) ^ "}"
    | print (CELL (P, NONE)) 
      = (printId P) ^ (if !mode = compressed then "_" else "[_]")
    | print (CELL (P, SOME P')) 
      = if !mode = compressed then print P' else 
        (printId P ^ "[" ^ print P' ^ "]")
    | print (VAR I) 
      = printId I
    | print (RATE r) 
      = r
    | print (ARC (P, (alpha, rate), Q)) 
      = (printId P) ^ "-(" ^ printId alpha ^ ", " ^ printRate rate ^ ")->" ^ (printId Q)
    | print (PRIORITIES p) 
      = printPriorities p
    | print (CONSTS (I, P, L)) 
      = (printId I)
   	 ^ (if !mode = verbose andalso notRate P then " =\n      " else " = ")
         ^ (print P) ^ ";\n" ^ (print L)
    | print (MARKING L) 
      = if !mode = compressed 
        then printMarking L 
        else "(" ^ printMarking L ^ ")"
  and printlist [] = ""
    | printlist [x] = print x
    | printlist (h::t) = print h ^ ", " ^ printlist t
  and printId s = (case !mode of 
		       compressed => HashTable.show s 
		     | uncompressed => HashTable.unhash s
		     | verbose => HashTable.unhash s)
  and printRate Top = "infty"
    | printRate (Rate r) = printId r
    | printRate (Product (i,r)) = Int.toString i ^ "*" ^ printId r
  and notRate (VAR I) = not (isNumeric I)
    | notRate _ = true
  and isNumeric s = List.all isNumericChar (explode (HashTable.unhash s))
  and isNumericChar c = Char.isDigit c orelse c = #"."
  and printPriorities nil  = ""
    | printPriorities [x]  = printList x
    | printPriorities (x::y::z)  = printList x ^ " > " ^ (printPriorities (y::z))
  and printList nil        = ""
    | printList [x]        = printId x
    | printList (x::y::z)  = printId x ^ ", " ^ (printList (y::z))
  and printMarking nil     = ""
    | printMarking [x]     = print x
    | printMarking (x::xs) = print x ^ (if !mode = compressed then "," else ",\n\t ") ^ 
                             printMarking xs

  fun pad s = s ^ "   "
  fun print' indent (PREFIX ((alpha, rate), P)) 
    = "("^printId alpha^", "^printRate rate^")."^(print' indent P)
    | print' indent (CHOICE (P, Q)) 
      = let val indent' = pad indent 
        in (print' indent' P) 
	   ^ ("\n     + ")
           ^ (print' indent' Q)
        end
    | print' indent (COOP ([P, Q], L)) 
      = let val indent' = pad indent 
        in "("
  	   ^ (print' indent' P) 
	   ^ (if L = [] then "\n" ^ indent ^ "<>\n" ^ indent  else 
              ("\n" ^ indent ^ "<\n" ^ indent' ^ (printList' indent' L) ^ 
               "\n" ^ indent ^ ">\n" ^ indent))
	   ^ (print' indent' Q) 
           ^ ")"
        end
    | print' indent (HIDING (P, [])) 
      = (print' indent P)
    | print' indent (HIDING (P, L)) 
      = let val indent' = pad indent 
        in (print' indent P) ^ "/{" ^ (printList' indent' L) ^ "}"
        end
    | print' indent (VAR I) 
      = printId I
    | print' indent P = print P
  and printList' indent nil        = ""
    | printList' indent [x]        = printId x
    | printList' indent (x::y::z)  = printId x ^ ",\n" ^ indent ^ 
                                     (printList' indent (y::z))
 
  val print = fn P => if !Options.compile then (print' "" P) ^ "\n" else print P
end;
