(* 
 File: Semantic.sml

 The structure which performs semantic analysis of PEPA models.
 Errors include component definitions which are missing or duplicated.
 Warnings include components which are defined but never used and
 occurrences of taus or wholly unused activity identifiers in any set 
 (hiding or cooperation).

*)


structure Semantic :> Semantic =
struct

  local
    (* Collects the set of all activity names *)
    fun act (PepaNets.PREFIX ((a, _), C)) = HashTable.unhash a :: act C
      | act (PepaNets.VAR _) = []
      | act (PepaNets.RATE _) = []
      | act (PepaNets.CHOICE (P, Q)) = act P @ act Q
      | act (PepaNets.COOP (P, _)) = List.concat (map act P)
      | act (PepaNets.HIDING (P, _)) = act P
      | act (PepaNets.CELL (P, NONE)) = []
      | act (PepaNets.CELL (P, SOME P')) = act P'
      | act (PepaNets.CONSTS (_, P, C)) = act P @ act C
      | act (PepaNets.ARC (P1, (a, _), P2)) = [HashTable.unhash a]
      | act (PepaNets.PRIORITIES _) = []
      | act (PepaNets.MARKING l) = List.concat (map act l)

    (* TO DO: Should check for redundant arc labels *)

    fun arcs (PepaNets.CONSTS (_, P, C)) = arcs P @ arcs C
      | arcs (PepaNets.ARC (_, (a, _), _)) = [HashTable.unhash a]
      | arcs _ = []

    fun priorities (PepaNets.PRIORITIES p) = map HashTable.unhash (List.concat p)
      | priorities (PepaNets.CONSTS (_, P, C)) = priorities P @ priorities C
      | priorities _ = []

    (* Collects the set of all component identifiers *)
    fun used (PepaNets.VAR I) = [HashTable.unhash I]
      | used (PepaNets.RATE _) = []
      | used (PepaNets.PREFIX (_, C)) = used C
      | used (PepaNets.CHOICE (P, Q)) = used P @ used Q
      | used (PepaNets.COOP (P, _)) = List.concat (map used P)
      | used (PepaNets.HIDING (P, _)) = used P
      | used (PepaNets.CELL (P, NONE)) = [HashTable.unhash P]
      | used (PepaNets.CELL (P, SOME P')) = HashTable.unhash P :: used P'
      | used (PepaNets.CONSTS (_, P, C)) = used P @ used C
      | used (PepaNets.ARC (P1, _, P2)) = map HashTable.unhash [P1, P2]
      | used (PepaNets.PRIORITIES _) = []
      | used (PepaNets.MARKING l) = List.concat (map used l)

    fun defined (PepaNets.CONSTS (I, PepaNets.RATE _, C)) = defined C
      | defined (PepaNets.CONSTS (I, _, C)) = HashTable.unhash I :: defined C
      | defined _ = []
  in
    val activities = Sort.quicksort o act
    val used = Sort.quicksort o used
    val defined = Sort.quicksort o defined
    val firings = Sort.quicksort o arcs
    val priorities = Sort.quicksort o priorities
  end

  fun rmDup [] = []
    | rmDup [x] = ([x] : string list)
    | rmDup (x1 :: (t as (x2 :: _))) =
        let val tail = rmDup t
         in if x1 = x2 then tail else x1 :: tail
        end

  local 
    fun fstNotSnd ([], _) = []
      | fstNotSnd (fst, []) = (fst : string list)
      | fstNotSnd (h1::fst, h2::snd) =
	if h1 < h2 
	then h1 :: fstNotSnd (fst, h2::snd)
	else if h1 = h2 then fstNotSnd (fst, h2::snd)
	     else fstNotSnd (h1::fst, snd)
  in
    fun firstNotSecond (fst, snd) = fstNotSnd (rmDup fst, rmDup snd)
  end

  fun checkDup [] = ()
    | checkDup [_] = ()
    | checkDup (x1 :: (t as (x2 :: _))) =
      if x1 = x2 andalso x1 <> "_ARC" andalso x1 <> "_PRIORITIES"
      then 
         Error.fatal_error ("Component multiply defined: " ^ x1)
      else checkDup t

  fun reportNotUsed C = 
      if C <> "_ARC" andalso C <> "_PRIORITIES" then
        Error.warning ("Component definition unused: " ^ C)
      else ()

  fun reportNotDefined C = 
      Error.fatal_error ("Component not defined: " ^ C)

  (* This version of member works on sorted string lists *)
  fun member (s : string) =
    let
      fun mem [] = false
        | mem (h::t) =
          not (h > s) andalso (h = s orelse mem t)          
    in mem end

  fun purifylist act class =
    let
      fun pL [] = []
        | pL (h :: t) = 
           case HashTable.unhash h of
             "tau" =>
               (Error.warning ("tau found in " ^ class ^ " set, ignoring");
                pL t)
           | id  => 
               if member id act
               then id :: pL t
               else (Error.warning ("unused activity name `" ^
                  id ^ "' found in " ^ class ^ " set, ignoring");
                pL t)

    in (map HashTable.hash) o pL end

  fun purify act =
    let 
      fun p (PepaNets.COOP (P, L)) = 
          PepaNets.COOP (map p P, purifylist act "cooperation" L)
        | p (PepaNets.HIDING (P, L)) =
          PepaNets.HIDING (p P, purifylist act "hiding" L)
        | p (PepaNets.CONSTS (I, P, C)) =
          PepaNets.CONSTS (I, p P, p C)
        | p C = C
    in p end

  fun priorityError p = 
      Error.fatal_error ("Priority set on non-firing: " ^ p)

  fun analyse C = 
    let
      val usedActivities = rmDup (activities C)
      val usedNames = used C
      val definedNames = defined C
      val defNotUsed = firstNotSecond (definedNames, usedNames)
      val _ = map reportNotUsed defNotUsed
      val usedNotDefined = firstNotSecond (usedNames, definedNames)
      val _ = map reportNotDefined usedNotDefined
      val _ = checkDup definedNames

      val prioritiesErrors = firstNotSecond (priorities C, firings C)
      val _ = map priorityError prioritiesErrors

    in
      purify usedActivities C
    end

end;
