(* 
 File: XMLPrettyprinter.sml

 The structure which prettyprints PEPA net models
 as XML output
*)

structure XMLPrettyprinter :> XMLPrettyprinter =
struct

  open PepaNets

  datatype printMode = verbose

  val mode = ref verbose

  fun setMode m = mode := m

  fun pad 0 s = s
    | pad n s = " " ^ (pad (n - 1) s)

  fun PEPAsyntax s Component = 
      let val _ = Prettyprinter.setMode Prettyprinter.verbose
      in
        s ^ "<PEPAsyntax>\n" ^
        s ^ "  " ^ "<![CDATA[\n" ^
        s ^ "    " ^ (Prettyprinter.print Component) ^ "\n" ^
        s ^ "  " ^ "]]>\n" ^
        s ^ "</PEPAsyntax>\n" 
      end

  fun print s (Component as (PREFIX ((alpha, rate), P)))
      = s ^ "<Component>\n" ^
                   PEPAsyntax (pad 2 s) Component ^
	s ^ "  " ^ "<Prefix>\n" ^ 
	s ^ "    " ^ "<Activity>\n" ^ 
			(printId' (pad 6 s) alpha) ^ 
			(printRate' (pad 6 s) rate) ^
	s ^ "    " ^ "</Activity>\n" ^ 
	                (print (pad 4 s) P) ^
	s ^ "  " ^ "</Prefix>\n" ^
	s ^ "</Component>\n"
    | print s (Component as (CHOICE (P, Q)))
      = s ^ "<Component>\n" ^ 
                   PEPAsyntax (pad 2 s) Component ^
        s ^ "  " ^ "<Choice>\n" ^ 
                      (print (pad 4 s) P) ^
                      (print (pad 4 s) Q) ^ 
        s ^ "  " ^ "</Choice>\n" ^ 
        s ^ "</Component>\n"
    | print s (Component as ((COOP ([], L))))
      = Error.fatal_error ("non-well-formed empty co-operation")
    | print s (Component as (COOP ([P, Q], L)))
      = s ^ "<Component>\n" ^ 
                   PEPAsyntax (pad 2 s) Component ^
        s ^ "  " ^ "<Cooperation>\n" ^ 
                      (print (pad 4 s) P) ^
                      (print (pad 4 s) Q) ^ 
        s ^ "    " ^ "<CooperationSet>\n" ^ 
                      (printList (pad 6 s) L) ^ 
        s ^ "    " ^ "</CooperationSet>\n" ^ 
        s ^ "  " ^ "</Cooperation>\n" ^ 
        s ^ "</Component>\n"
    | print s (Component as (COOP (Ps, L)))
      = s ^ "<Component>\n" ^ 
                   PEPAsyntax (pad 2 s) Component ^
        s ^ "  " ^ "<Cooperation>\n" ^ 
                      (printlist (pad 4 s) Ps) ^
        s ^ "    " ^ "<CooperationSet>\n" ^ 
                      (printList (pad 6 s) L) ^ 
        s ^ "    " ^ "</CooperationSet>\n" ^ 
        s ^ "  " ^ "</Cooperation>\n" ^ 
        s ^ "</Component>\n"
    | print s (Component as (HIDING (P, [])))
      = (print s P)
    | print s (Component as (HIDING (P, L)))
      = s ^ "<Component>\n" ^ 
                   PEPAsyntax (pad 2 s) Component ^
        s ^ "  " ^ "<Hiding>\n" ^ 
                      (print (pad 4 s) P) ^
                      (printList (pad 4 s) L) ^ 
        s ^ "  " ^ "</Hiding>\n" ^ 
        s ^ "</Component>\n"
    | print s (Component as (CELL (P, NONE)))
      = s ^ "<Component>\n" ^ 
                   PEPAsyntax (pad 2 s) Component ^
        s ^ "  " ^ "<Cell>\n" ^ 
                      (printId' (pad 4 s) P) ^ 
        s ^ "    " ^ "<Contents />\n" ^ 
        s ^ "  " ^ "</Cell>\n" ^ 
        s ^ "</Component>\n"
    | print s (Component as (CELL (P, SOME P')))
      = s ^ "<Component>\n" ^ 
                   PEPAsyntax (pad 2 s) Component ^
        s ^ "  " ^ "<Cell>\n" ^ 
                      (printId' (pad 4 s) P) ^ 
        s ^ "    " ^ "<Contents>\n" ^ 
                      (print (pad 6 s) P') ^ 
        s ^ "    " ^ "</Contents>\n" ^ 
        s ^ "  " ^ "</Cell>\n" ^ 
        s ^ "</Component>\n"
    | print s (VAR I) 
      = s ^ "<Component>\n" ^ 
        s ^ "  " ^ "<Identifier>" ^ (printId I) ^ "</Identifier>\n" ^
        s ^ "</Component>\n"
    | print s (RATE r) 
      = s ^ "<Rate>" ^ r ^ "</Rate>\n"
    | print s (ARC (P, (alpha, rate), Q)) 
      = s ^ "<Arc>\n" ^
        s ^ "  " ^ "<InputPlace>" ^ (printId P) ^ "</InputPlace>\n" ^ 
        s ^ "  " ^ "<ArcLabel>\n" ^ 
        s ^ "    " ^ "<ArcLabelActivityType>" ^ (printId alpha) ^ 
                     "</ArcLabelActivityType>\n" ^ 
        s ^ "    " ^ "<ArcLabelActivityRate>" ^ (printRate rate) ^ 
                     "</ArcLabelActivityRate>\n" ^ 
        s ^ "  " ^ "<OutputPlace>" ^ (printId Q) ^ "</OutputPlace>\n" ^ 
        s ^ "</Arc>\n" 
    | print s (PRIORITIES p) 
      = s ^ "<Priorities>\n" ^
              printPriorities s p ^
        s ^ "</Priorities>\n" 
    | print s (CONSTS (I, P, L)) 
      = s ^ "<Constant>\n" ^
        s ^ "  " ^ "<ConstantIdentifier>" ^ (printId I) ^ "</ConstantIdentifier>\n" ^
                    (print (pad 2 s) P) ^ 
        s ^ "</Constant>\n" ^
             (print s L)
    | print s (Component as (MARKING L))
      = s ^ "<Marking>\n" ^
                PEPAsyntax (pad 2 s) Component ^
                printMarking (pad 2 s)  L ^
        s ^ "</Marking>\n" 
  and printlist s [] = ""
    | printlist s [x] = print s x
    | printlist s (h::t) = print s h ^ printlist s t
  and printId id = HashTable.unhash id
  and printRate Top = "infty"
    | printRate (Rate r) = printId r
    | printRate (Product (i,r)) = Int.toString i ^ "*" ^ printId r
  and notRate (VAR I) = not (isNumeric I)
    | notRate _ = true
  and isNumeric s = List.all isNumericChar (explode (HashTable.unhash s))
  and isNumericChar c = Char.isDigit c orelse c = #"."
  and printPriorities s nil  = ""
    | printPriorities s [x]  = printList s x
    | printPriorities s (x::y::z)  = 
      printList s x ^ s ^ "<NextPriority />\n" ^ (printPriorities s (y::z))
  and printList s nil        = ""
    | printList s [x]        = printId' s  x
    | printList s (x::y::z)  = printId' s  x ^ ", " ^ (printList s (y::z))
  and printId' s i = s ^ "<Id>" ^ HashTable.unhash i ^ "</Id>\n"
  and printRate' s r = s ^ "<Rate>" ^ printRate r ^ "</Rate>\n"
  and printMarking s nil     = ""
    | printMarking s [x]     = print s x
    | printMarking s (x::xs) = print s x ^ printMarking s xs

  val print = fn P => print "  " P

end;


